use std::{env, fs, path::PathBuf, process, time::SystemTime};
use anyhow::Result as anyResult;
use clap::Parser;

#[derive(Parser)]
#[command(version = "0.1.5", about = "It's not a dust pan, it's a rust pan. (A cargo cleaner)", long_about = None)]
struct Args {
    #[arg(short='v', long="verbose")]
    /// Print directory even if it's not being cleaned
    verbose: bool,
    #[arg(short='d', long="days", default_value_t=30)]
    /// Days since last modified and/or last accessed
    days: u32,
    #[arg(short='f', long="force")]
    /// Clean the directory, ignoring days since last modified and/or last accessed
    force: bool,
}

fn find_target_dir(path: PathBuf) -> anyResult<bool> {
    let mut found_target_dir: bool = false;
    let mut found_cargo_toml: bool = false;
    for entry in fs::read_dir(path)? {
        let entry = entry?;
        let p = entry.path();
        if p.is_dir() && p.ends_with("target") {
            found_target_dir = true;
        } else if p.is_file() && p.file_name().unwrap() == "Cargo.toml" {
            found_cargo_toml = true;
        }
    }
    Ok(found_target_dir && found_cargo_toml)
}

fn find_newest_file(path: PathBuf) -> anyResult<PathBuf> {
    let mut pat: PathBuf = path.clone();
    let mut pat_dif: SystemTime = SystemTime::now();
    for entry in fs::read_dir(&path)? {
        let entry = entry?;
        let p = entry.path();
        if !p.is_file() {
            continue;
        }
        let meta = entry.metadata()?;
        let modif = meta.modified()?;
        let access = meta.accessed()?;
        if pat_dif.elapsed()?.as_secs() == 0 {
            pat = path.join(p);
            if modif.ge(&access) {
                pat_dif = modif;
            } else {
                pat_dif = access;
            }
        } else if pat_dif.le(&modif) || pat_dif.le(&access) {
            pat = path.join(p);
            if modif.ge(&access) {
                pat_dif = modif;
            } else {
                pat_dif = access;
            }
        }
    }
    Ok(pat)
}

fn execute_clean(path: PathBuf) -> anyResult<()> {
    let _cmd = process::Command::new("cargo")
        .current_dir(path)
        .arg("clean")
        .status()?;
    Ok(())
}

fn clean_in_dir(path: PathBuf, verbose: bool, forced: bool, days: u64) -> anyResult<u64> {
    let mut cleaned: u64 = 0;
    for entry in fs::read_dir(path)? {
        let entry = entry?;
        let p = entry.path();
        if p.is_dir() {
            let needed = find_target_dir(p.clone())?;
            if needed {
                if !forced {
                    // If forced no need to check
                    let newest = find_newest_file(p.clone())?;
                    let meta = fs::metadata(newest)?;
                    let modif = meta.modified()?;
                    let access = meta.accessed()?;
                    let el: u64;
                    if modif.ge(&access) {
                        el = modif.elapsed()?.as_secs();
                    } else {
                        el = access.elapsed()?.as_secs();
                    }
                    let days = el / 86400;
                    if days < days {
                        if verbose {
                            println!(" => {} (dirty, {} days)", p.display(), days);
                        }
                        continue;
                    }
                }
                if !verbose {
                    println!(" => {}", p.display());
                } else {
                    println!(" => {} (dirty)", p.display());
                }
                execute_clean(p)?;
                cleaned += 1;
            } else {
                // Maybe nested?
                let r = clean_in_dir(p, verbose, forced, days)?;
                cleaned += r;
            }
        }
    }
    Ok(cleaned)
}

fn main() -> anyResult<()> {
    let cli = Args::parse();
    println!("Sweeping...");
    let cwd = env::current_dir()?;
    let cleaned = clean_in_dir(cwd, cli.verbose, cli.force, cli.days as u64)?;
    if cleaned == 1 {
        println!("Cleaned 1 crate");
    } else {
        println!("Cleaned {} crates", cleaned);
    }
    Ok(())
}
